@@ -1,9 +1,10 @@
-import hello_pb2_grpc
-import hello_pb2
-import grpc
-from concurrent import futures
 import logging
 import time
+from concurrent import futures
+
+import grpc
+import hello_pb2
+import hello_pb2_grpc
 
 logger = logging.getLogger(__name__)
 
@@ -22,3 +23,7 @@ def start_server():
     server.start()
     logger.info("gRPC server started")
     server.wait_for_termination()
+
+
+if __name__ == "__main__":
+    start_server()
