@@ -1,8 +1,8 @@
-import os
-
-from locust import HttpUser, TaskSet, task, between
+from locust import HttpUser, TaskSet, between, task
 from locust.clients import HttpSession
 
+import os
+
 
 class MultipleHostsUser(HttpUser):
     abstract = True
