@@ -1,7 +1,7 @@
 #!/usr/bin/env python3
 
-import subprocess
 import os
+import subprocess
 import sys
 
 github_api_token = (
@@ -22,9 +22,9 @@ cmd = [
     "-p",
     "locust",
     "--exclude-labels",
-    "duplicate,question,invalid,wontfix,cantfix,stale",
+    "duplicate,question,invalid,wontfix,cantfix,stale,no-changelog",
     "--header-label",
-    "# Changelog\nNote that the latest version is usually missing here, check [github](https://github.com/locustio/locust/releases) for the latest info.",
+    "# Detailed changelog\nThe most important changes can also be found in [the documentation](https://docs.locust.io/en/latest/changelog.html).",
     "--since-tag",
     "1.0.1",
     # "--since-commit", # these cause issues
