@@ -1,4 +1,9 @@
+from __future__ import annotations
+
+from locust import __version__
+
 import functools
+import inspect
 import json
 import logging
 import os
@@ -17,34 +22,23 @@ from operator import (
 from types import TracebackType
 from typing import (
     TYPE_CHECKING,
-    Dict,
-    Iterator,
-    List,
-    NoReturn,
-    ValuesView,
-    Set,
-    Optional,
-    Tuple,
-    Type,
     Any,
+    Callable,
+    Iterator,
+    NoReturn,
+    TypedDict,
+    ValuesView,
     cast,
 )
 from uuid import uuid4
 
-# @TODO: typing.Protocol is in python >= 3.8
-try:
-    from typing import Protocol, TypedDict
-except ImportError:
-    from typing_extensions import Protocol, TypedDict  # type: ignore
-
 import gevent
 import greenlet
 import psutil
 from gevent.event import Event
 from gevent.pool import Group
 
-from . import User
-from locust import __version__
+from . import argument_parser
 from .dispatch import UsersDispatcher
 from .exception import RPCError, RPCReceiveError, RPCSendError
 from .log import greenlet_exception_logger
@@ -57,9 +51,9 @@ from .stats import (
     StatsError,
     setup_distributed_stats_event_listeners,
 )
-from . import argument_parser
 
 if TYPE_CHECKING:
+    from . import User
     from .env import Environment
 
 logger = logging.getLogger(__name__)
@@ -79,6 +73,7 @@ CPU_WARNING_THRESHOLD = 90
 HEARTBEAT_INTERVAL = 1
 HEARTBEAT_LIVENESS = 3
 HEARTBEAT_DEAD_INTERNAL = -60
+MASTER_HEARTBEAT_TIMEOUT = 60
 FALLBACK_INTERVAL = 5
 CONNECT_TIMEOUT = 5
 CONNECT_RETRY_COUNT = 60
@@ -90,13 +85,7 @@ class ExceptionDict(TypedDict):
     count: int
     msg: str
     traceback: str
-    nodes: Set[str]
-
-
-class CustomMessageListener(Protocol):
-    @abstractmethod
-    def __call__(self, environment: "Environment", msg: Message) -> None:
-        ...
+    nodes: set[str]
 
 
 class Runner:
@@ -110,47 +99,40 @@ class Runner:
     desired type.
     """
 
-    def __init__(self, environment: "Environment") -> None:
+    def __init__(self, environment: Environment) -> None:
         self.environment = environment
         self.user_greenlets = Group()
         self.greenlet = Group()
         self.state = STATE_INIT
-        self.spawning_greenlet: Optional[gevent.Greenlet] = None
-        self.shape_greenlet: Optional[gevent.Greenlet] = None
-        self.shape_last_tick: Tuple[int, float] | Tuple[int, float, Optional[List[Type[User]]]] | None = None
+        self.spawning_greenlet: gevent.Greenlet | None = None
+        self.shape_greenlet: gevent.Greenlet | None = None
+        self.shape_last_tick: tuple[int, float] | tuple[int, float, list[type[User]] | None] | None = None
         self.current_cpu_usage: int = 0
         self.cpu_warning_emitted: bool = False
         self.worker_cpu_warning_emitted: bool = False
         self.current_memory_usage: int = 0
         self.greenlet.spawn(self.monitor_cpu_and_memory).link_exception(greenlet_exception_handler)
-        self.exceptions: Dict[int, ExceptionDict] = {}
+        self.exceptions: dict[int, ExceptionDict] = {}
         # Because of the way the ramp-up/ramp-down is implemented, target_user_classes_count
         # is only updated at the end of the ramp-up/ramp-down.
         # See https://github.com/locustio/locust/issues/1883#issuecomment-919239824 for context.
-        self.target_user_classes_count: Dict[str, int] = {}
+        self.target_user_classes_count: dict[str, int] = {}
         # target_user_count is set before the ramp-up/ramp-down occurs.
         self.target_user_count: int = 0
-        self.custom_messages: Dict[str, CustomMessageListener] = {}
+        self.custom_messages: dict[str, Callable] = {}
 
-        self._users_dispatcher: Optional[UsersDispatcher] = None
+        self._users_dispatcher: UsersDispatcher | None = None
 
         # set up event listeners for recording requests
-        def on_request_success(request_type, name, response_time, response_length, **_kwargs):
+        def on_request(request_type, name, response_time, response_length, exception=None, **_kwargs):
             self.stats.log_request(request_type, name, response_time, response_length)
+            if exception:
+                self.stats.log_error(request_type, name, exception)
 
-        def on_request_failure(request_type, name, response_time, response_length, exception, **_kwargs):
-            self.stats.log_request(request_type, name, response_time, response_length)
-            self.stats.log_error(request_type, name, exception)
-
-        # temporarily set log level to ignore warnings to suppress deprecation message
-        loglevel = logging.getLogger().level
-        logging.getLogger().setLevel(logging.ERROR)
-        self.environment.events.request_success.add_listener(on_request_success)
-        self.environment.events.request_failure.add_listener(on_request_failure)
-        logging.getLogger().setLevel(loglevel)
+        self.environment.events.request.add_listener(on_request)
 
         self.connection_broken = False
-        self.final_user_classes_count: Dict[str, int] = {}  # just for the ratio report, fills before runner stops
+        self.final_user_classes_count: dict[str, int] = {}  # just for the ratio report, fills before runner stops
 
         # register listener that resets stats when spawning is complete
         def on_spawning_complete(user_count: int) -> None:
@@ -167,11 +149,11 @@ class Runner:
             self.greenlet.kill(block=False)
 
     @property
-    def user_classes(self) -> List[Type[User]]:
+    def user_classes(self) -> list[type[User]]:
         return self.environment.user_classes
 
     @property
-    def user_classes_by_name(self) -> Dict[str, Type[User]]:
+    def user_classes_by_name(self) -> dict[str, type[User]]:
         return self.environment.user_classes_by_name
 
     @property
@@ -179,7 +161,7 @@ class Runner:
         return self.environment.stats
 
     @property
-    def errors(self) -> Dict[str, StatsError]:
+    def errors(self) -> dict[str, StatsError]:
         return self.stats.errors
 
     @property
@@ -190,7 +172,7 @@ class Runner:
         return len(self.user_greenlets)
 
     @property
-    def user_classes_count(self) -> Dict[str, int]:
+    def user_classes_count(self) -> dict[str, int]:
         """
         :returns: Number of currently running users for each user class
         """
@@ -230,18 +212,17 @@ class Runner:
             )
         return self.cpu_warning_emitted
 
-    def spawn_users(self, user_classes_spawn_count: Dict[str, int], wait: bool = False):
+    def spawn_users(self, user_classes_spawn_count: dict[str, int], wait: bool = False):
         if self.state == STATE_INIT or self.state == STATE_STOPPED:
             self.update_state(STATE_SPAWNING)
 
         logger.debug(
-            "Spawning additional %s (%s already running)..."
-            % (json.dumps(user_classes_spawn_count), json.dumps(self.user_classes_count))
+            f"Spawning additional {json.dumps(user_classes_spawn_count)} ({json.dumps(self.user_classes_count)} already running)..."
         )
 
-        def spawn(user_class: str, spawn_count: int) -> List[User]:
+        def spawn(user_class: str, spawn_count: int) -> list[User]:
             n = 0
-            new_users: List[User] = []
+            new_users: list[User] = []
             while n < spawn_count:
                 new_user = self.user_classes_by_name[user_class](self.environment)
                 new_user.start(self.user_greenlets)
@@ -249,10 +230,10 @@ class Runner:
                 n += 1
                 if n % 10 == 0 or n == spawn_count:
                     logger.debug("%i users spawned" % self.user_count)
-            logger.debug(f"All users of class {user_class} spawned")
+            logger.debug("All users of class %s spawned" % user_class)
             return new_users
 
-        new_users: List[User] = []
+        new_users: list[User] = []
         for user_class, spawn_count in user_classes_spawn_count.items():
             new_users += spawn(user_class, spawn_count)
 
@@ -261,7 +242,7 @@ class Runner:
             logger.info("All users stopped\n")
         return new_users
 
-    def stop_users(self, user_classes_stop_count: Dict[str, int]) -> None:
+    def stop_users(self, user_classes_stop_count: dict[str, int]) -> None:
         async_calls_to_stop = Group()
         stop_group = Group()
 
@@ -269,7 +250,7 @@ class Runner:
             if self.user_classes_count[user_class] == 0:
                 continue
 
-            to_stop: List[greenlet.greenlet] = []
+            to_stop: list[greenlet.greenlet] = []
             for user_greenlet in self.user_greenlets:
                 if len(to_stop) == stop_count:
                     break
@@ -288,7 +269,7 @@ class Runner:
 
             while True:
                 user_to_stop: User = to_stop.pop()
-                logger.debug(f"Stopping {user_to_stop.greenlet.name}")
+                logger.debug("Stopping %s" % user_to_stop.greenlet.name)
                 if user_to_stop.greenlet is greenlet.getcurrent():
                     # User called runner.quit(), so don't block waiting for killing to finish
                     user_to_stop.group.killone(user_to_stop.greenlet, block=False)
@@ -329,10 +310,14 @@ class Runner:
 
     @abstractmethod
     def start(
-        self, user_count: int, spawn_rate: float, wait: bool = False, user_classes: Optional[List[Type[User]]] = None
+        self, user_count: int, spawn_rate: float, wait: bool = False, user_classes: list[type[User]] | None = None
     ) -> None:
         ...
 
+    @abstractmethod
+    def send_message(self, msg_type: str, data: Any | None = None, client_id: str | None = None) -> None:
+        ...
+
     def start_shape(self) -> None:
         """
         Start running a load test with a custom LoadTestShape specified in the :meth:`Environment.shape_class <locust.env.Environment.shape_class>` parameter.
@@ -341,7 +326,7 @@ class Runner:
             logger.info("There is an ongoing shape test running. Editing is disabled")
             return
 
-        logger.info("Shape test starting. User count and spawn rate are ignored for this type of load test")
+        logger.info("Shape test starting.")
         self.update_state(STATE_INIT)
         self.shape_greenlet = self.greenlet.spawn(self.shape_worker)
         self.shape_greenlet.link_exception(greenlet_exception_handler)
@@ -351,6 +336,7 @@ class Runner:
     def shape_worker(self) -> None:
         logger.info("Shape worker starting")
         while self.state == STATE_INIT or self.state == STATE_SPAWNING or self.state == STATE_RUNNING:
+            shape_adjustment_start = time.time()
             current_tick = self.environment.shape_class.tick() if self.environment.shape_class is not None else None
             if current_tick is None:
                 logger.info("Shape test stopping")
@@ -361,9 +347,7 @@ class Runner:
                 self.shape_greenlet = None
                 self.shape_last_tick = None
                 return
-            elif self.shape_last_tick == current_tick:
-                gevent.sleep(1)
-            else:
+            elif self.shape_last_tick != current_tick:
                 if len(current_tick) == 2:
                     user_count, spawn_rate = current_tick  # type: ignore
                     user_classes = None
@@ -385,6 +369,8 @@ class Runner:
                 #        of each load test shape stage.
                 self.start(user_count=user_count, spawn_rate=spawn_rate, user_classes=user_classes)
                 self.shape_last_tick = current_tick
+            shape_adjustment_time_ms = time.time() - shape_adjustment_start
+            gevent.sleep(max(1 - shape_adjustment_time_ms, 0))
 
     def stop(self) -> None:
         """
@@ -392,7 +378,11 @@ class Runner:
         """
         if self.state == STATE_STOPPED:
             return
-        logger.debug("Stopping all users")
+        try:
+            caller = inspect.getframeinfo(inspect.stack()[1][0])
+            logger.debug(f"Stopping all users (called from {caller.filename}:{caller.lineno})")
+        except Exception:
+            logger.debug("Stopping all users (couldnt determine where stop() was called from)")
         self.environment.events.test_stopping.fire(environment=self.environment)
         self.final_user_classes_count = {**self.user_classes_count}
         self.update_state(STATE_CLEANUP)
@@ -432,13 +422,15 @@ class Runner:
         row["nodes"].add(node_id)
         self.exceptions[key] = row
 
-    def register_message(self, msg_type: str, listener: CustomMessageListener) -> None:
+    def register_message(self, msg_type: str, listener: Callable) -> None:
         """
         Register a listener for a custom message from another node
 
         :param msg_type: The type of the message to listen for
         :param listener: The function to execute when the message is received
         """
+        if msg_type in self.custom_messages:
+            raise Exception(f"Tried to register listener method for {msg_type}, but it already had a listener!")
         self.custom_messages[msg_type] = listener
 
 
@@ -447,15 +439,15 @@ class LocalRunner(Runner):
     Runner for running single process load test
     """
 
-    # always set to 0 for LocalRunner
-    worker_index = 0
-
     def __init__(self, environment) -> None:
         """
         :param environment: Environment instance
         """
         super().__init__(environment)
-
+        # These attributes dont make a lot of sense for LocalRunner
+        # but it makes it easier to write tests that work for both local and distributed runs
+        self.worker_index = 0
+        self.client_id = socket.gethostname() + "_" + uuid4().hex
         # Only when running in standalone mode (non-distributed)
         self._local_worker_node = WorkerNode(id="local")
         self._local_worker_node.user_classes_count = self.user_classes_count
@@ -467,7 +459,7 @@ class LocalRunner(Runner):
 
         self.environment.events.user_error.add_listener(on_user_error)
 
-    def _start(self, user_count: int, spawn_rate: float, wait: bool = False, user_classes: list = None) -> None:
+    def _start(self, user_count: int, spawn_rate: float, wait: bool = False, user_classes: list | None = None) -> None:
         """
         Start running a load test
 
@@ -500,20 +492,20 @@ class LocalRunner(Runner):
             self.update_state(STATE_SPAWNING)
 
         if self._users_dispatcher is None:
-            self._users_dispatcher = UsersDispatcher(
+            self._users_dispatcher = self.environment.dispatcher_class(
                 worker_nodes=[self._local_worker_node], user_classes=self.user_classes
             )
 
         logger.info("Ramping to %d users at a rate of %.2f per second" % (user_count, spawn_rate))
 
-        cast(UsersDispatcher, self._users_dispatcher).new_dispatch(user_count, spawn_rate, user_classes)
+        self._users_dispatcher.new_dispatch(user_count, spawn_rate, user_classes)
 
         try:
             for dispatched_users in self._users_dispatcher:
-                user_classes_spawn_count: Dict[str, int] = {}
-                user_classes_stop_count: Dict[str, int] = {}
+                user_classes_spawn_count: dict[str, int] = {}
+                user_classes_stop_count: dict[str, int] = {}
                 user_classes_count = dispatched_users[self._local_worker_node.id]
-                logger.debug(f"Ramping to {_format_user_classes_count_for_log(user_classes_count)}")
+                logger.debug("Ramping to %s" % _format_user_classes_count_for_log(user_classes_count))
                 for user_class_name, user_class_count in user_classes_count.items():
                     if self.user_classes_count[user_class_name] > user_class_count:
                         user_classes_stop_count[user_class_name] = (
@@ -549,7 +541,7 @@ class LocalRunner(Runner):
         self.environment.events.spawning_complete.fire(user_count=sum(self.target_user_classes_count.values()))
 
     def start(
-        self, user_count: int, spawn_rate: float, wait: bool = False, user_classes: Optional[List[Type[User]]] = None
+        self, user_count: int, spawn_rate: float, wait: bool = False, user_classes: list[type[User]] | None = None
     ) -> None:
         if spawn_rate > 100:
             logger.warning(
@@ -569,14 +561,14 @@ class LocalRunner(Runner):
             return
         super().stop()
 
-    def send_message(self, msg_type: str, data: Optional[Any] = None) -> None:
+    def send_message(self, msg_type: str, data: Any | None = None, client_id: str | None = None) -> None:
         """
         Emulates internodal messaging by calling registered listeners
 
         :param msg_type: The type of the message to emulate sending
         :param data: Optional data to include
         """
-        logger.debug(f"Running locally: sending {msg_type} message to self")
+        logger.debug("Running locally: sending %s message to self" % msg_type)
         if msg_type in self.custom_messages:
             listener = self.custom_messages[msg_type]
             msg = Message(msg_type, data, "local")
@@ -600,7 +592,7 @@ class WorkerNode:
         self.cpu_warning_emitted = False
         self.memory_usage: int = 0
         # The reported users running on the worker
-        self.user_classes_count: Dict[str, int] = {}
+        self.user_classes_count: dict[str, int] = {}
 
     @property
     def user_count(self) -> int:
@@ -609,9 +601,9 @@ class WorkerNode:
 
 class WorkerNodes(MutableMapping):
     def __init__(self):
-        self._worker_nodes: Dict[str, WorkerNode] = {}
+        self._worker_nodes: dict[str, WorkerNode] = {}
 
-    def get_by_state(self, state) -> List[WorkerNode]:
+    def get_by_state(self, state) -> list[WorkerNode]:
         return [c for c in self.values() if c.state == state]
 
     @property
@@ -619,19 +611,19 @@ class WorkerNodes(MutableMapping):
         return self.values()
 
     @property
-    def ready(self) -> List[WorkerNode]:
+    def ready(self) -> list[WorkerNode]:
         return self.get_by_state(STATE_INIT)
 
     @property
-    def spawning(self) -> List[WorkerNode]:
+    def spawning(self) -> list[WorkerNode]:
         return self.get_by_state(STATE_SPAWNING)
 
     @property
-    def running(self) -> List[WorkerNode]:
+    def running(self) -> list[WorkerNode]:
         return self.get_by_state(STATE_RUNNING)
 
     @property
-    def missing(self) -> List[WorkerNode]:
+    def missing(self) -> list[WorkerNode]:
         return self.get_by_state(STATE_MISSING)
 
     def __setitem__(self, k: str, v: WorkerNode) -> None:
@@ -690,13 +682,13 @@ class MasterRunner(DistributedRunner):
             else:
                 raise
 
-        self._users_dispatcher: Optional[UsersDispatcher] = None
+        self._users_dispatcher: UsersDispatcher | None = None
 
         self.greenlet.spawn(self.heartbeat_worker).link_exception(greenlet_exception_handler)
         self.greenlet.spawn(self.client_listener).link_exception(greenlet_exception_handler)
 
         # listener that gathers info on how many users the worker has spawned
-        def on_worker_report(client_id: str, data: Dict[str, Any]) -> None:
+        def on_worker_report(client_id: str, data: dict[str, Any]) -> None:
             if client_id not in self.clients:
                 logger.info("Discarded report from unrecognized worker %s", client_id)
                 return
@@ -705,7 +697,7 @@ class MasterRunner(DistributedRunner):
         self.environment.events.worker_report.add_listener(on_worker_report)
 
         # register listener that sends quit message to worker nodes
-        def on_quitting(environment: "Environment", **kw):
+        def on_quitting(environment: Environment, **kw):
             self.quit()
 
         self.environment.events.quitting.add_listener(on_quitting)
@@ -740,7 +732,7 @@ class MasterRunner(DistributedRunner):
         return warning_emitted
 
     def start(
-        self, user_count: int, spawn_rate: float, wait=False, user_classes: Optional[List[Type[User]]] = None
+        self, user_count: int, spawn_rate: float, wait=False, user_classes: list[type[User]] | None = None
     ) -> None:
         self.spawning_completed = False
 
@@ -758,7 +750,7 @@ class MasterRunner(DistributedRunner):
         self.spawn_rate = spawn_rate
 
         if self._users_dispatcher is None:
-            self._users_dispatcher = UsersDispatcher(
+            self._users_dispatcher = self.environment.dispatcher_class(
                 worker_nodes=list(self.clients.values()), user_classes=self.user_classes
             )
 
@@ -848,7 +840,7 @@ class MasterRunner(DistributedRunner):
 
         logger.info(f"{msg_prefix}: {_format_user_classes_count_for_log(self.reported_user_classes_count)}")
 
-    @functools.lru_cache()
+    @functools.lru_cache
     def _wait_for_workers_report_after_ramp_up(self) -> float:
         """
         The amount of time to wait after a ramp-up in order for all the workers to report their state
@@ -892,11 +884,11 @@ class MasterRunner(DistributedRunner):
 
             if send_stop_to_client:
                 for client in self.clients.all:
-                    logger.debug(f"Sending stop message to worker {client.id}")
+                    logger.debug("Sending stop message to worker %s" % client.id)
                     self.server.send_to_client(Message("stop", None, client.id))
 
                 # Give an additional 60s for all workers to stop
-                timeout = gevent.Timeout(self.environment.stop_timeout or 0 + 60)
+                timeout = gevent.Timeout(self.environment.stop_timeout + 60)
                 timeout.start()
                 try:
                     while self.user_count != 0:
@@ -918,25 +910,9 @@ class MasterRunner(DistributedRunner):
 
     def check_stopped(self) -> None:
         if (
-            not self.state == STATE_INIT
-            and not self.state == STATE_STOPPED
-            and (
-                (
-                    self.state == STATE_STOPPING
-                    and all(
-                        map(
-                            lambda x: x.state == STATE_INIT,
-                            self.clients.all,
-                        )
-                    )
-                )
-            )
-            or all(
-                map(
-                    lambda x: x.state not in (STATE_RUNNING, STATE_SPAWNING, STATE_INIT),
-                    self.clients.all,
-                )
-            )
+            self.state == STATE_STOPPING
+            and all(x.state == STATE_INIT for x in self.clients.all)
+            or all(x.state not in (STATE_RUNNING, STATE_SPAWNING, STATE_INIT) for x in self.clients.all)
         ):
             self.update_state(STATE_STOPPED)
 
@@ -993,13 +969,20 @@ class MasterRunner(DistributedRunner):
             try:
                 client_id, msg = self.server.recv_from_client()
             except RPCReceiveError as e:
-                logger.error(f"RPCError when receiving from client: {e}. Will reset client {client_id}.")
-                try:
-                    self.server.send_to_client(Message("reconnect", None, client_id))
-                except Exception as e:
-                    logger.error(f"Error sending reconnect message to worker: {e}. Will reset RPC server.")
-                    self.connection_broken = True
-                    gevent.sleep(FALLBACK_INTERVAL)
+                client_id = e.addr
+
+                if client_id and client_id in self.clients:
+                    logger.error(f"RPCError when receiving from client: {e}. Will reset client {client_id}.")
+                    try:
+                        self.server.send_to_client(Message("reconnect", None, client_id))
+                    except Exception as error:
+                        logger.error(f"Error sending reconnect message to worker: {error}. Will reset RPC server.")
+                        self.connection_broken = True
+                        gevent.sleep(FALLBACK_INTERVAL)
+                        continue
+                else:
+                    message = f"{e}" if not client_id else f"{e} from {client_id}"
+                    logger.error(f"Unrecognized message detected: {message}")
                     continue
             except RPCSendError as e:
                 logger.error(f"Error sending reconnect message to worker: {e}. Will reset RPC server.")
@@ -1017,6 +1000,10 @@ class MasterRunner(DistributedRunner):
                 self.connection_broken = True
                 gevent.sleep(FALLBACK_INTERVAL)
                 continue
+            except KeyboardInterrupt:
+                logging.debug(
+                    "Got KeyboardInterrupt in client_listener. Other greenlets should catch this and shut down."
+                )
             if msg.type == "client_ready":
                 if not msg.data:
                     logger.error(f"An old (pre 2.0) worker tried to connect ({client_id}). That's not going to work.")
@@ -1031,7 +1018,7 @@ class MasterRunner(DistributedRunner):
                             f"A worker ({client_id}) running a different version ({msg.data}) connected, master version is {__version__}"
                         )
                 self.send_message("ack", client_id=client_id, data={"index": self.get_worker_index(client_id)})
-
+                self.environment.events.worker_connect.fire(client_id=msg.node_id)
                 self.clients[client_id] = WorkerNode(client_id, heartbeat_liveness=HEARTBEAT_LIVENESS)
                 if self._users_dispatcher is not None:
                     self._users_dispatcher.add_worker(worker_node=self.clients[client_id])
@@ -1046,6 +1033,40 @@ class MasterRunner(DistributedRunner):
                 # emit a warning if the worker's clock seem to be out of sync with our clock
                 # if abs(time() - msg.data["time"]) > 5.0:
                 #    warnings.warn("The worker node's clock seem to be out of sync. For the statistics to be correct the different locust servers need to have synchronized clocks.")
+            elif msg.type == "locustfile":
+                logging.debug("Worker requested locust file")
+                assert self.environment.parsed_options
+                filename = (
+                    "locustfile.py"
+                    if self.environment.parsed_options.locustfile == "locustfile"
+                    else self.environment.parsed_options.locustfile
+                )
+                try:
+                    with open(filename) as f:
+                        file_contents = f.read()
+                except Exception as e:
+                    logger.error(
+                        f"--locustfile must be a plain filename (not a module name) for file distribution to work {e}"
+                    )
+                    self.send_message(
+                        "locustfile",
+                        client_id=client_id,
+                        data={
+                            "error": f"locustfile parameter on master must be a plain filename (not a module name) (was '{filename}')"
+                        },
+                    )
+                else:
+                    if getattr(self, "_old_file_contents", file_contents) != file_contents:
+                        logger.warning(
+                            "Locustfile contents changed on disk after first worker requested locustfile, sending new content. If you make any major changes (like changing User class names) you need to restart master."
+                        )
+                    self._old_file_contents = file_contents
+                    self.send_message(
+                        "locustfile",
+                        client_id=client_id,
+                        data={"filename": os.path.basename(filename), "contents": file_contents},
+                    )
+                continue
             elif msg.type == "client_stopped":
                 if msg.node_id not in self.clients:
                     logger.warning(f"Received {msg.type} message from an unknown worker: {msg.node_id}.")
@@ -1082,10 +1103,17 @@ class MasterRunner(DistributedRunner):
                         )
                     if "current_memory_usage" in msg.data:
                         c.memory_usage = msg.data["current_memory_usage"]
+                    self.server.send_to_client(Message("heartbeat", None, msg.node_id))
+                else:
+                    logging.debug(f"Got heartbeat message from unknown worker {msg.node_id}")
             elif msg.type == "stats":
                 self.environment.events.worker_report.fire(client_id=msg.node_id, data=msg.data)
             elif msg.type == "spawning":
-                self.clients[msg.node_id].state = STATE_SPAWNING
+                try:
+                    self.clients[msg.node_id].state = STATE_SPAWNING
+                except KeyError:
+                    logger.warning(f"Got spawning message from unknown worker {msg.node_id}. Asking worker to quit.")
+                    self.server.send_to_client(Message("quit", None, msg.node_id))
             elif msg.type == "spawning_complete":
                 self.clients[msg.node_id].state = STATE_RUNNING
                 self.clients[msg.node_id].user_classes_count = msg.data["user_classes_count"]
@@ -1112,7 +1140,11 @@ class MasterRunner(DistributedRunner):
                 logger.debug(
                     f"Received {msg.type} message from worker {msg.node_id} (index {self.get_worker_index(msg.node_id)})"
                 )
-                self.custom_messages[msg.type](environment=self.environment, msg=msg)
+                try:
+                    self.custom_messages[msg.type](environment=self.environment, msg=msg)
+                except Exception:
+                    logging.error(f"Uncaught exception in handler for {msg.type}\n{traceback.format_exc()}")
+
             else:
                 logger.warning(
                     f"Unknown message type received from worker {msg.node_id} (index {self.get_worker_index(msg.node_id)}): {msg.type}"
@@ -1125,14 +1157,14 @@ class MasterRunner(DistributedRunner):
         return len(self.clients.ready) + len(self.clients.spawning) + len(self.clients.running)
 
     @property
-    def reported_user_classes_count(self) -> Dict[str, int]:
-        reported_user_classes_count: Dict[str, int] = defaultdict(lambda: 0)
+    def reported_user_classes_count(self) -> dict[str, int]:
+        reported_user_classes_count: dict[str, int] = defaultdict(int)
         for client in self.clients.ready + self.clients.spawning + self.clients.running:
             for name, count in client.user_classes_count.items():
                 reported_user_classes_count[name] += count
         return reported_user_classes_count
 
-    def send_message(self, msg_type: str, data: Optional[Dict[str, Any]] = None, client_id: Optional[str] = None):
+    def send_message(self, msg_type: str, data: dict[str, Any] | None = None, client_id: str | None = None):
         """
         Sends a message to attached worker node(s)
 
@@ -1162,7 +1194,7 @@ class WorkerRunner(DistributedRunner):
     # the worker index is set on ACK, if master provided it (masters <= 2.10.2 do not provide it)
     worker_index = -1
 
-    def __init__(self, environment: "Environment", master_host: str, master_port: int) -> None:
+    def __init__(self, environment: Environment, master_host: str, master_port: int) -> None:
         """
         :param environment: Environment instance
         :param master_host: Host/IP to use for connection to the master
@@ -1171,17 +1203,19 @@ class WorkerRunner(DistributedRunner):
         super().__init__(environment)
         self.retry = 0
         self.connected = False
+        self.last_heartbeat_timestamp: float | None = None
         self.connection_event = Event()
         self.worker_state = STATE_INIT
         self.client_id = socket.gethostname() + "_" + uuid4().hex
         self.master_host = master_host
         self.master_port = master_port
         self.worker_cpu_warning_emitted = False
-        self._users_dispatcher: Optional[UsersDispatcher] = None
+        self._users_dispatcher: UsersDispatcher | None = None
         self.client = rpc.Client(master_host, master_port, self.client_id)
         self.greenlet.spawn(self.worker).link_exception(greenlet_exception_handler)
         self.connect_to_master()
         self.greenlet.spawn(self.heartbeat).link_exception(greenlet_exception_handler)
+        self.greenlet.spawn(self.heartbeat_timeout_checker).link_exception(greenlet_exception_handler)
         self.greenlet.spawn(self.stats_reporter).link_exception(greenlet_exception_handler)
 
         # register listener for when all users have spawned, and report it to the master node
@@ -1199,14 +1233,14 @@ class WorkerRunner(DistributedRunner):
         self.environment.events.spawning_complete.add_listener(on_spawning_complete)
 
         # register listener that adds the current number of spawned users to the report that is sent to the master node
-        def on_report_to_master(client_id: str, data: Dict[str, Any]):
+        def on_report_to_master(client_id: str, data: dict[str, Any]):
             data["user_classes_count"] = self.user_classes_count
             data["user_count"] = self.user_count
 
         self.environment.events.report_to_master.add_listener(on_report_to_master)
 
         # register listener that sends quit message to master
-        def on_quitting(environment: "Environment", **kw) -> None:
+        def on_quitting(environment: Environment, **kw) -> None:
             self.client.send(Message("quit", None, self.client_id))
 
         self.environment.events.quitting.add_listener(on_quitting)
@@ -1219,11 +1253,11 @@ class WorkerRunner(DistributedRunner):
         self.environment.events.user_error.add_listener(on_user_error)
 
     def start(
-        self, user_count: int, spawn_rate: float, wait: bool = False, user_classes: Optional[List[Type[User]]] = None
+        self, user_count: int, spawn_rate: float, wait: bool = False, user_classes: list[type[User]] | None = None
     ) -> None:
         raise NotImplementedError("use start_worker")
 
-    def start_worker(self, user_classes_count: Dict[str, int], **kwargs) -> None:
+    def start_worker(self, user_classes_count: dict[str, int], **kwargs) -> None:
         """
         Start running a load test as a worker
 
@@ -1236,8 +1270,8 @@ class WorkerRunner(DistributedRunner):
             if self.environment.host:
                 user_class.host = self.environment.host
 
-        user_classes_spawn_count: Dict[str, int] = {}
-        user_classes_stop_count: Dict[str, int] = {}
+        user_classes_spawn_count: dict[str, int] = {}
+        user_classes_stop_count: dict[str, int] = {}
 
         for user_class_name, user_class_count in user_classes_count.items():
             if self.user_classes_count[user_class_name] > user_class_count:
@@ -1271,6 +1305,13 @@ class WorkerRunner(DistributedRunner):
                 self.reset_connection()
             gevent.sleep(HEARTBEAT_INTERVAL)
 
+    def heartbeat_timeout_checker(self) -> NoReturn:
+        while True:
+            gevent.sleep(1)
+            if self.last_heartbeat_timestamp and self.last_heartbeat_timestamp < time.time() - MASTER_HEARTBEAT_TIMEOUT:
+                logger.error(f"Didn't get heartbeat from master in over {MASTER_HEARTBEAT_TIMEOUT}s")
+                self.quit()
+
     def reset_connection(self) -> None:
         logger.info("Reset connection to master")
         try:
@@ -1301,7 +1342,7 @@ class WorkerRunner(DistributedRunner):
                     )
                     continue
                 self.environment.host = job["host"]
-                self.environment.stop_timeout = job["stop_timeout"]
+                self.environment.stop_timeout = job["stop_timeout"] or 0.0
 
                 # receive custom arguments
                 if self.environment.parsed_options is None:
@@ -1339,7 +1380,7 @@ class WorkerRunner(DistributedRunner):
                 # +additional_wait is just a small buffer to account for the random network latencies and/or other
                 # random delays inherent to distributed systems.
                 additional_wait = int(os.getenv("LOCUST_WORKER_ADDITIONAL_WAIT_BEFORE_READY_AFTER_STOP", 0))
-                gevent.sleep((self.environment.stop_timeout or 0) + additional_wait)
+                gevent.sleep(self.environment.stop_timeout + additional_wait)
                 self.client.send(Message("client_ready", __version__, self.client_id))
                 self.worker_state = STATE_INIT
             elif msg.type == "quit":
@@ -1350,8 +1391,10 @@ class WorkerRunner(DistributedRunner):
             elif msg.type == "reconnect":
                 logger.warning("Received reconnect message from master. Resetting RPC connection.")
                 self.reset_connection()
+            elif msg.type == "heartbeat":
+                self.last_heartbeat_timestamp = time.time()
             elif msg.type in self.custom_messages:
-                logger.debug(f"Received {msg.type} message from master")
+                logger.debug("Received %s message from master" % msg.type)
                 self.custom_messages[msg.type](environment=self.environment, msg=msg)
             else:
                 logger.warning(f"Unknown message type received: {msg.type}")
@@ -1364,25 +1407,30 @@ class WorkerRunner(DistributedRunner):
                 logger.error(f"Temporary connection lost to master server: {e}, will retry later.")
             gevent.sleep(WORKER_REPORT_INTERVAL)
 
-    def send_message(self, msg_type: str, data: Optional[Dict[str, Any]] = None) -> None:
+    def send_message(self, msg_type: str, data: dict[str, Any] | None = None, client_id: str | None = None) -> None:
         """
         Sends a message to master node
 
         :param msg_type: The type of the message to send
         :param data: Optional data to send
+        :param client_id: (unused)
         """
-        logger.debug(f"Sending {msg_type} message to master")
+        logger.debug("Sending %s message to master" % msg_type)
         self.client.send(Message(msg_type, data, self.client_id))
 
     def _send_stats(self) -> None:
-        data: Dict[str, Any] = {}
+        data: dict[str, Any] = {}
         self.environment.events.report_to_master.fire(client_id=self.client_id, data=data)
         self.client.send(Message("stats", data, self.client_id))
 
     def connect_to_master(self):
         self.retry += 1
         self.client.send(Message("client_ready", __version__, self.client_id))
-        success = self.connection_event.wait(timeout=CONNECT_TIMEOUT)
+        try:
+            success = self.connection_event.wait(timeout=CONNECT_TIMEOUT)
+        except KeyboardInterrupt:
+            # dont complain about getting CTRL-C
+            sys.exit(1)
         if not success:
             if self.retry < 3:
                 logger.debug(
@@ -1398,14 +1446,14 @@ class WorkerRunner(DistributedRunner):
         self.connected = True
 
 
-def _format_user_classes_count_for_log(user_classes_count: Dict[str, int]) -> str:
+def _format_user_classes_count_for_log(user_classes_count: dict[str, int]) -> str:
     return "{} ({} total users)".format(
         json.dumps(dict(sorted(user_classes_count.items(), key=itemgetter(0)))),
         sum(user_classes_count.values()),
     )
 
 
-def _aggregate_dispatched_users(d: Dict[str, Dict[str, int]]) -> Dict[str, int]:
+def _aggregate_dispatched_users(d: dict[str, dict[str, int]]) -> dict[str, int]:
     # TODO: Test it
     user_classes = list(next(iter(d.values())).keys())
     return {u: sum(d[u] for d in d.values()) for u in user_classes}
