@@ -1,48 +1,44 @@
 from __future__ import annotations
-from abc import abstractmethod
+
+import csv
 import datetime
 import hashlib
-from tempfile import NamedTemporaryFile
-import time
-from collections import namedtuple, OrderedDict
-from copy import copy
-from itertools import chain
+import json
+import logging
 import os
-import csv
 import signal
-import gevent
-
+import time
+from abc import abstractmethod
+from collections import OrderedDict, defaultdict, namedtuple
+from copy import copy
+from html import escape
+from itertools import chain
+from tempfile import NamedTemporaryFile
+from types import FrameType
 from typing import (
     TYPE_CHECKING,
     Any,
-    Dict,
+    Callable,
     Iterable,
     NoReturn,
-    Tuple,
-    List,
-    Optional,
-    OrderedDict as OrderedDictType,
-    Callable,
+    Protocol,
+    TypedDict,
     TypeVar,
     cast,
 )
+from typing import (
+    OrderedDict as OrderedDictType,
+)
 
-# @TODO: typing.Protocol is in python >= 3.8
-try:
-    from typing import Protocol, TypedDict
-except ImportError:
-    from typing_extensions import Protocol, TypedDict  # type: ignore
+import gevent
 
-from types import FrameType
-
-from .exception import CatchResponseError
 from .event import Events
-
-import logging
+from .exception import CatchResponseError
+from .util.rounding import proper_round
 
 if TYPE_CHECKING:
-    from .runners import Runner
     from .env import Environment
+    from .runners import Runner
 
 console_logger = logging.getLogger("locust.stats_logger")
 
@@ -67,18 +63,18 @@ class StatsBaseDict(TypedDict):
 
 
 class StatsEntryDict(StatsBaseDict):
-    last_request_timestamp: Optional[float]
+    last_request_timestamp: float | None
     start_time: float
     num_requests: int
     num_none_requests: int
     num_failures: int
     total_response_time: int
     max_response_time: int
-    min_response_time: Optional[int]
+    min_response_time: int | None
     total_content_length: int
-    response_times: Dict[int, int]
-    num_reqs_per_sec: Dict[int, int]
-    num_fail_per_sec: Dict[int, int]
+    response_times: dict[int, int]
+    num_reqs_per_sec: dict[int, int]
+    num_fail_per_sec: dict[int, int]
 
 
 class StatsErrorDict(StatsBaseDict):
@@ -94,7 +90,7 @@ class StatsHolder(Protocol):
 S = TypeVar("S", bound=StatsHolder)
 
 
-def resize_handler(signum: int, frame: Optional[FrameType]):
+def resize_handler(signum: int, frame: FrameType | None):
     global STATS_NAME_WIDTH
     if STATS_AUTORESIZE:
         try:
@@ -120,24 +116,26 @@ HISTORY_STATS_INTERVAL_SEC = 5
 CSV_STATS_INTERVAL_SEC = 1
 CSV_STATS_FLUSH_INTERVAL_SEC = 10
 
-
 """
 Default window size/resolution - in seconds - when calculating the current
 response time percentile
 """
 CURRENT_RESPONSE_TIME_PERCENTILE_WINDOW = 10
 
-
 CachedResponseTimes = namedtuple("CachedResponseTimes", ["response_times", "num_requests"])
 
 PERCENTILES_TO_REPORT = [0.50, 0.66, 0.75, 0.80, 0.90, 0.95, 0.98, 0.99, 0.999, 0.9999, 1.0]
 
+PERCENTILES_TO_STATISTICS = [0.95, 0.99]
+PERCENTILES_TO_CHART = [0.50, 0.95]
+MODERN_UI_PERCENTILES_TO_CHART = [0.95]
+
 
 class RequestStatsAdditionError(Exception):
     pass
 
 
-def get_readable_percentiles(percentile_list: List[float]) -> List[str]:
+def get_readable_percentiles(percentile_list: list[float]) -> list[str]:
     """
     Converts a list of percentiles from 0-1 fraction to 0%-100% view for using in console & csv reporting
     :param percentile_list: The list of percentiles in range 0-1
@@ -149,7 +147,7 @@ def get_readable_percentiles(percentile_list: List[float]) -> List[str]:
     ]
 
 
-def calculate_response_time_percentile(response_times: Dict[int, int], num_requests: int, percent: float) -> int:
+def calculate_response_time_percentile(response_times: dict[int, int], num_requests: int, percent: float) -> int:
     """
     Get the response time that a certain number of percent of the requests
     finished within. Arguments:
@@ -170,7 +168,7 @@ def calculate_response_time_percentile(response_times: Dict[int, int], num_reque
     return 0
 
 
-def diff_response_time_dicts(latest: Dict[int, int], old: Dict[int, int]) -> Dict[int, int]:
+def diff_response_time_dicts(latest: dict[int, int], old: dict[int, int]) -> dict[int, int]:
     """
     Returns the delta between two {response_times:request_count} dicts.
 
@@ -186,6 +184,17 @@ def diff_response_time_dicts(latest: Dict[int, int], old: Dict[int, int]) -> Dic
     return new
 
 
+class EntriesDict(dict):
+    def __init__(self, request_stats):
+        self.request_stats = request_stats
+
+    def __missing__(self, key):
+        self[key] = StatsEntry(
+            self.request_stats, key[0], key[1], use_response_times_cache=self.request_stats.use_response_times_cache
+        )
+        return self[key]
+
+
 class RequestStats:
     """
     Class that holds the request statistics. Accessible in a User from self.environment.stats
@@ -199,8 +208,8 @@ class RequestStats:
                                          is not needed.
         """
         self.use_response_times_cache = use_response_times_cache
-        self.entries: Dict[Tuple[str, str], StatsEntry] = {}
-        self.errors: Dict[str, StatsError] = {}
+        self.entries: dict[tuple[str, str], StatsEntry] = EntriesDict(self)
+        self.errors: dict[str, StatsError] = {}
         self.total = StatsEntry(self, "Aggregated", None, use_response_times_cache=self.use_response_times_cache)
         self.history = []
 
@@ -226,11 +235,11 @@ class RequestStats:
 
     def log_request(self, method: str, name: str, response_time: int, content_length: int) -> None:
         self.total.log(response_time, content_length)
-        self.get(name, method).log(response_time, content_length)
+        self.entries[(name, method)].log(response_time, content_length)
 
     def log_error(self, method: str, name: str, error: Exception | str | None) -> None:
         self.total.log_error(error)
-        self.get(name, method).log_error(error)
+        self.entries[(name, method)].log_error(error)
 
         # store error in errors dict
         key = StatsError.create_key(method, name, error)
@@ -240,15 +249,11 @@ class RequestStats:
             self.errors[key] = entry
         entry.occurred()
 
-    def get(self, name: str, method: str) -> "StatsEntry":
+    def get(self, name: str, method: str) -> StatsEntry:
         """
         Retrieve a StatsEntry instance by name and method
         """
-        entry = self.entries.get((name, method))
-        if not entry:
-            entry = StatsEntry(self, name, method, use_response_times_cache=self.use_response_times_cache)
-            self.entries[(name, method)] = entry
-        return entry
+        return self.entries[(name, method)]
 
     def reset_all(self) -> None:
         """
@@ -265,18 +270,16 @@ class RequestStats:
         Remove all stats entries and errors
         """
         self.total = StatsEntry(self, "Aggregated", "", use_response_times_cache=self.use_response_times_cache)
-        self.entries = {}
+        self.entries = EntriesDict(self)
         self.errors = {}
         self.history = []
 
-    def serialize_stats(self) -> List["StatsEntryDict"]:
+    def serialize_stats(self) -> list[StatsEntryDict]:
         return [
-            self.entries[key].get_stripped_report()
-            for key in self.entries.keys()
-            if not (self.entries[key].num_requests == 0 and self.entries[key].num_failures == 0)
+            e.get_stripped_report() for e in self.entries.values() if not (e.num_requests == 0 and e.num_failures == 0)
         ]
 
-    def serialize_errors(self) -> Dict[str, "StatsErrorDict"]:
+    def serialize_errors(self) -> dict[str, StatsErrorDict]:
         return {k: e.serialize() for k, e in self.errors.items()}
 
 
@@ -285,7 +288,7 @@ class StatsEntry:
     Represents a single stats entry (name and method)
     """
 
-    def __init__(self, stats: Optional[RequestStats], name: str, method: str, use_response_times_cache: bool = False):
+    def __init__(self, stats: RequestStats | None, name: str, method: str, use_response_times_cache: bool = False):
         self.stats = stats
         self.name = name
         """ Name (URL) of this stats entry """
@@ -306,15 +309,15 @@ class StatsEntry:
         """ Number of failed request """
         self.total_response_time: int = 0
         """ Total sum of the response times """
-        self.min_response_time: Optional[int] = None
+        self.min_response_time: int | None = None
         """ Minimum response time """
         self.max_response_time: int = 0
         """ Maximum response time """
-        self.num_reqs_per_sec: Dict[int, int] = {}
+        self.num_reqs_per_sec: dict[int, int] = defaultdict(int)
         """ A {second => request_count} dict that holds the number of requests made per second """
-        self.num_fail_per_sec: Dict[int, int] = {}
+        self.num_fail_per_sec: dict[int, int] = defaultdict(int)
         """ A (second => failure_count) dict that hold the number of failures per second """
-        self.response_times: Dict[int, int] = {}
+        self.response_times: dict[int, int] = defaultdict(int)
         """
         A {response_time => count} dict that holds the response time distribution of all
         the requests.
@@ -324,16 +327,16 @@ class StatsEntry:
 
         This dict is used to calculate the median and percentile response times.
         """
-        self.response_times_cache: Optional[OrderedDictType[int, CachedResponseTimes]] = None
+        self.response_times_cache: OrderedDictType[int, CachedResponseTimes] | None = None
         """
         If use_response_times_cache is set to True, this will be a {timestamp => CachedResponseTimes()}
         OrderedDict that holds a copy of the response_times dict for each of the last 20 seconds.
         """
         self.total_content_length: int = 0
-        """ The sum of the content length of all the requests for this entry """
+        """ The sum of the content length of all the responses for this entry """
         self.start_time: float = 0.0
         """ Time of the first request for this entry """
-        self.last_request_timestamp: Optional[float] = None
+        self.last_request_timestamp: float | None = None
         """ Time of the last request for this entry """
         self.reset()
 
@@ -343,12 +346,12 @@ class StatsEntry:
         self.num_none_requests = 0
         self.num_failures = 0
         self.total_response_time = 0
-        self.response_times = {}
+        self.response_times = defaultdict(int)
         self.min_response_time = None
         self.max_response_time = 0
         self.last_request_timestamp = None
-        self.num_reqs_per_sec = {}
-        self.num_fail_per_sec = {}
+        self.num_reqs_per_sec = defaultdict(int)
+        self.num_fail_per_sec = defaultdict(int)
         self.total_content_length = 0
         if self.use_response_times_cache:
             self.response_times_cache = OrderedDict()
@@ -372,7 +375,7 @@ class StatsEntry:
 
     def _log_time_of_request(self, current_time: float) -> None:
         t = int(current_time)
-        self.num_reqs_per_sec[t] = self.num_reqs_per_sec.setdefault(t, 0) + 1
+        self.num_reqs_per_sec[t] += 1
         self.last_request_timestamp = current_time
 
     def _log_response_time(self, response_time: int) -> None:
@@ -384,8 +387,8 @@ class StatsEntry:
 
         if self.min_response_time is None:
             self.min_response_time = response_time
-
-        self.min_response_time = min(self.min_response_time, response_time)
+        else:
+            self.min_response_time = min(self.min_response_time, response_time)
         self.max_response_time = max(self.max_response_time, response_time)
 
         # to avoid to much data that has to be transferred to the master node when
@@ -401,13 +404,12 @@ class StatsEntry:
             rounded_response_time = round(response_time, -3)
 
         # increase request count for the rounded key in response time dict
-        self.response_times.setdefault(rounded_response_time, 0)
         self.response_times[rounded_response_time] += 1
 
     def log_error(self, error: Exception | str | None) -> None:
         self.num_failures += 1
         t = int(time.time())
-        self.num_fail_per_sec[t] = self.num_fail_per_sec.setdefault(t, 0) + 1
+        self.num_fail_per_sec[t] += 1
 
     @property
     def fail_ratio(self) -> float:
@@ -449,7 +451,7 @@ class StatsEntry:
             return 0
         slice_start_time = max(int(self.stats.last_request_timestamp) - 12, int(self.stats.start_time or 0))
 
-        reqs: List[int | float] = [
+        reqs: list[int | float] = [
             self.num_reqs_per_sec.get(t, 0) for t in range(slice_start_time, int(self.stats.last_request_timestamp) - 2)
         ]
         return avg(reqs)
@@ -490,7 +492,7 @@ class StatsEntry:
         except ZeroDivisionError:
             return 0
 
-    def extend(self, other: "StatsEntry") -> None:
+    def extend(self, other: StatsEntry) -> None:
         """
         Extend the data from the current StatsEntry with the stats from another
         StatsEntry instance.
@@ -538,7 +540,7 @@ class StatsEntry:
         return cast(StatsEntryDict, {key: getattr(self, key, None) for key in StatsEntryDict.__annotations__.keys()})
 
     @classmethod
-    def unserialize(cls, data: StatsEntryDict) -> "StatsEntry":
+    def unserialize(cls, data: StatsEntryDict) -> StatsEntry:
         """Return the unserialzed version of the specified dict"""
         obj = cls(None, data["name"], data["method"])
         valid_keys = StatsEntryDict.__annotations__.keys()
@@ -601,7 +603,7 @@ class StatsEntry:
         """
         return calculate_response_time_percentile(self.response_times, self.num_requests, percent)
 
-    def get_current_response_time_percentile(self, percent: float) -> Optional[int]:
+    def get_current_response_time_percentile(self, percent: float) -> int | None:
         """
         Calculate the *current* response time for a certain percentile. We use a sliding
         window of (approximately) the last 10 seconds (specified by CURRENT_RESPONSE_TIME_PERCENTILE_WINDOW)
@@ -619,13 +621,13 @@ class StatsEntry:
         # when trying to fetch the cached response_times. We construct this list in such a way
         # that it's ordered by preference by starting to add t-10, then t-11, t-9, t-12, t-8,
         # and so on
-        acceptable_timestamps: List[int] = []
+        acceptable_timestamps: list[int] = []
         acceptable_timestamps.append(t - CURRENT_RESPONSE_TIME_PERCENTILE_WINDOW)
         for i in range(1, 9):
             acceptable_timestamps.append(t - CURRENT_RESPONSE_TIME_PERCENTILE_WINDOW - i)
             acceptable_timestamps.append(t - CURRENT_RESPONSE_TIME_PERCENTILE_WINDOW + i)
 
-        cached: Optional[CachedResponseTimes] = None
+        cached: CachedResponseTimes | None = None
         if self.response_times_cache is not None:
             for ts in acceptable_timestamps:
                 if ts in self.response_times_cache:
@@ -677,6 +679,30 @@ class StatsEntry:
             for _ in range(len(self.response_times_cache) - cache_size):
                 self.response_times_cache.popitem(last=False)
 
+    def to_dict(self, escape_string_values=False):
+        response_time_percentiles = {
+            f"response_time_percentile_{percentile}": self.get_response_time_percentile(percentile)
+            for percentile in PERCENTILES_TO_STATISTICS
+        }
+
+        return {
+            "method": escape(self.method or "") if escape_string_values else self.method,
+            "name": escape(self.name) if escape_string_values else self.name,
+            "safe_name": escape(self.name, quote=False),
+            "num_requests": self.num_requests,
+            "num_failures": self.num_failures,
+            "avg_response_time": self.avg_response_time,
+            "min_response_time": 0 if self.min_response_time is None else proper_round(self.min_response_time),
+            "max_response_time": proper_round(self.max_response_time),
+            "current_rps": self.current_rps,
+            "current_fail_per_sec": self.current_fail_per_sec,
+            "median_response_time": self.median_response_time,
+            "ninetieth_response_time": self.get_response_time_percentile(0.9),  # for legacy ui
+            "ninety_ninth_response_time": self.get_response_time_percentile(0.99),  # for legacy ui
+            **response_time_percentiles,  # for modern ui
+            "avg_content_length": self.avg_content_length,
+        }
+
 
 class StatsError:
     def __init__(self, method: str, name: str, error: Exception | str | None, occurrences: int = 0):
@@ -723,7 +749,7 @@ class StatsError:
         return f"{self.method} {self.name}: {unwrapped_error}"
 
     def serialize(self) -> StatsErrorDict:
-        def _getattr(obj: "StatsError", key: str, default: Optional[Any]) -> Optional[Any]:
+        def _getattr(obj: StatsError, key: str, default: Any | None) -> Any | None:
             value = getattr(obj, key, default)
 
             if key in ["error"]:
@@ -734,15 +760,23 @@ class StatsError:
         return cast(StatsErrorDict, {key: _getattr(self, key, None) for key in StatsErrorDict.__annotations__.keys()})
 
     @classmethod
-    def unserialize(cls, data: StatsErrorDict) -> "StatsError":
+    def unserialize(cls, data: StatsErrorDict) -> StatsError:
         return cls(data["method"], data["name"], data["error"], data["occurrences"])
 
+    def to_dict(self, escape_string_values=False):
+        return {
+            "method": escape(self.method),
+            "name": escape(self.name),
+            "error": escape(self.parse_error(self.error)),
+            "occurrences": self.occurrences,
+        }
 
-def avg(values: List[float | int]) -> float:
+
+def avg(values: list[float | int]) -> float:
     return sum(values, 0.0) / max(len(values), 1)
 
 
-def median_from_dict(total: int, count: Dict[int, int]) -> int:
+def median_from_dict(total: int, count: dict[int, int]) -> int:
     """
     total is the number of requests made
     count is a dict {response_time: count}
@@ -757,13 +791,13 @@ def median_from_dict(total: int, count: Dict[int, int]) -> int:
 
 
 def setup_distributed_stats_event_listeners(events: Events, stats: RequestStats) -> None:
-    def on_report_to_master(client_id: str, data: Dict[str, Any]) -> None:
+    def on_report_to_master(client_id: str, data: dict[str, Any]) -> None:
         data["stats"] = stats.serialize_stats()
         data["stats_total"] = stats.total.get_stripped_report()
         data["errors"] = stats.serialize_errors()
         stats.errors = {}
 
-    def on_worker_report(client_id: str, data: Dict[str, Any]) -> None:
+    def on_worker_report(client_id: str, data: dict[str, Any]) -> None:
         for stats_data in data["stats"]:
             entry = StatsEntry.unserialize(stats_data)
             request_key = (entry.name, entry.method)
@@ -789,7 +823,11 @@ def print_stats(stats: RequestStats, current=True) -> None:
     console_logger.info("")
 
 
-def get_stats_summary(stats: RequestStats, current=True) -> List[str]:
+def print_stats_json(stats: RequestStats) -> None:
+    print(json.dumps(stats.serialize_stats(), indent=4))
+
+
+def get_stats_summary(stats: RequestStats, current=True) -> list[str]:
     """
     stats summary will be returned as list of string
     """
@@ -815,7 +853,7 @@ def print_percentile_stats(stats: RequestStats) -> None:
     console_logger.info("")
 
 
-def get_percentile_stats_summary(stats: RequestStats) -> List[str]:
+def get_percentile_stats_summary(stats: RequestStats) -> list[str]:
     """
     Percentile stats summary will be returned as list of string
     """
@@ -844,13 +882,12 @@ def get_percentile_stats_summary(stats: RequestStats) -> List[str]:
 
 
 def print_error_report(stats: RequestStats) -> None:
-    if not len(stats.errors):
-        return
-    for line in get_error_report_summary(stats):
-        console_logger.info(line)
+    if stats.errors:
+        for line in get_error_report_summary(stats):
+            console_logger.info(line)
 
 
-def get_error_report_summary(stats) -> List[str]:
+def get_error_report_summary(stats) -> list[str]:
     summary = ["Error report"]
     summary.append("%-18s %-100s" % ("# occurrences", "Error"))
     separator = f'{"-" * 18}|{"-" * ((80 + STATS_NAME_WIDTH) - 19)}'
@@ -871,23 +908,33 @@ def stats_printer(stats: RequestStats) -> Callable[[], None]:
     return stats_printer_func
 
 
-def sort_stats(stats: Dict[Any, S]) -> List[S]:
+def sort_stats(stats: dict[Any, S]) -> list[S]:
     return [stats[key] for key in sorted(stats.keys())]
 
 
-def stats_history(runner: "Runner") -> None:
+def stats_history(runner: Runner) -> None:
     """Save current stats info to history for charts of report."""
     while True:
         stats = runner.stats
         if not stats.total.use_response_times_cache:
             break
         if runner.state != "stopped":
+            current_response_time_percentiles = {
+                f"response_time_percentile_{percentile}": stats.total.get_current_response_time_percentile(percentile)
+                or 0
+                for percentile in MODERN_UI_PERCENTILES_TO_CHART
+            }
+
             r = {
-                "time": datetime.datetime.utcnow().strftime("%H:%M:%S"),
+                **current_response_time_percentiles,
+                "time": datetime.datetime.now(tz=datetime.timezone.utc).strftime("%H:%M:%S"),
                 "current_rps": stats.total.current_rps or 0,
                 "current_fail_per_sec": stats.total.current_fail_per_sec or 0,
-                "response_time_percentile_95": stats.total.get_current_response_time_percentile(0.95) or 0,
-                "response_time_percentile_50": stats.total.get_current_response_time_percentile(0.5) or 0,
+                "response_time_percentile_1": stats.total.get_current_response_time_percentile(PERCENTILES_TO_CHART[0])
+                or 0,
+                "response_time_percentile_2": stats.total.get_current_response_time_percentile(PERCENTILES_TO_CHART[1])
+                or 0,
+                "total_avg_response_time": stats.total.avg_response_time,
                 "user_count": runner.user_count or 0,
             }
             stats.history.append(r)
@@ -897,7 +944,7 @@ def stats_history(runner: "Runner") -> None:
 class StatsCSV:
     """Write statistics to csv_writer stream."""
 
-    def __init__(self, environment: "Environment", percentiles_to_report: List[float]) -> None:
+    def __init__(self, environment: Environment, percentiles_to_report: list[float]) -> None:
         self.environment = environment
         self.percentiles_to_report = percentiles_to_report
 
@@ -931,7 +978,7 @@ class StatsCSV:
             "Nodes",
         ]
 
-    def _percentile_fields(self, stats_entry: StatsEntry, use_current: bool = False) -> List[str] | List[int]:
+    def _percentile_fields(self, stats_entry: StatsEntry, use_current: bool = False) -> list[str] | list[int]:
         if not stats_entry.num_requests:
             return self.percentiles_na
         elif use_current:
@@ -999,8 +1046,8 @@ class StatsCSVFileWriter(StatsCSV):
 
     def __init__(
         self,
-        environment: "Environment",
-        percentiles_to_report: List[float],
+        environment: Environment,
+        percentiles_to_report: list[float],
         base_filepath: str,
         full_history: bool = False,
     ):
@@ -1096,7 +1143,7 @@ class StatsCSVFileWriter(StatsCSV):
 
         stats = self.environment.stats
         timestamp = int(now)
-        stats_entries: List[StatsEntry] = []
+        stats_entries: list[StatsEntry] = []
         if self.full_history:
             stats_entries = sort_stats(stats.entries)
 
