@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import logging
 
 from typing import List, Union, Dict
@@ -7,11 +9,11 @@ LogMessage = List[Union[str, Dict[str, TracebackType]]]
 
 
 class MockedLoggingHandler(logging.Handler):
-    debug: List[LogMessage] = []
-    warning: List[LogMessage] = []
-    info: List[LogMessage] = []
-    error: List[LogMessage] = []
-    critical: List[LogMessage] = []
+    debug: list[LogMessage] = []
+    warning: list[LogMessage] = []
+    info: list[LogMessage] = []
+    error: list[LogMessage] = []
+    critical: list[LogMessage] = []
 
     def emit(self, record):
         if record.exc_info:
