@@ -1,19 +1,20 @@
-import unittest
-import os
-from tempfile import NamedTemporaryFile, TemporaryDirectory
-from random import randint
-from unittest import mock
-from io import StringIO
-
 import locust
 from locust.argument_parser import (
-    locustfile_is_directory,
-    parse_options,
-    get_parser,
-    parse_locustfile_option,
-    ui_extra_args_dict,
     find_locustfiles,
+    get_parser,
+    locustfile_is_directory,
+    parse_locustfile_option,
+    parse_options,
+    ui_extra_args_dict,
 )
+
+import os
+import unittest
+from io import StringIO
+from random import randint
+from tempfile import NamedTemporaryFile, TemporaryDirectory
+from unittest import mock
+
 from .mock_locustfile import mock_locustfile
 from .testcases import LocustTestCase
 
@@ -37,22 +38,50 @@ class TestParser(unittest.TestCase):
         opts = self.parser.parse_args(args)
         self.assertEqual(opts.skip_log_setup, True)
 
-    def test_parameter_parsing(self):
-        with NamedTemporaryFile(mode="w") as file:
-            os.environ["LOCUST_LOCUSTFILE"] = "locustfile_from_env"
-            file.write("host host_from_config\nweb-host webhost_from_config")
+    def test_parse_options_from_conf_file(self):
+        with NamedTemporaryFile(mode="w", suffix=".conf") as file:
+            config_data = """\
+            locustfile = ./test_locustfile.py
+            web-host = 127.0.0.1
+            web-port = 45787
+            headless
+            tags = [Critical, Normal]
+            """
+
+            file.write(config_data)
             file.flush()
             parser = get_parser(default_config_files=[file.name])
-            options = parser.parse_args(["-H", "host_from_args"])
-            del os.environ["LOCUST_LOCUSTFILE"]
-        self.assertEqual(options.web_host, "webhost_from_config")
-        self.assertEqual(options.locustfile, "locustfile_from_env")
-        self.assertEqual(options.host, "host_from_args")  # overridden
+            options = parser.parse_args(["-H", "https://example.com"])
 
-    def test_web_auth(self):
-        args = ["--web-auth", "hello:bye"]
-        opts = self.parser.parse_args(args)
-        self.assertEqual(opts.web_auth, "hello:bye")
+        self.assertEqual("./test_locustfile.py", options.locustfile)
+        self.assertEqual("127.0.0.1", options.web_host)
+        self.assertEqual(45787, options.web_port)
+        self.assertTrue(options.headless)
+        self.assertEqual(["Critical", "Normal"], options.tags)
+        self.assertEqual("https://example.com", options.host)
+
+    def test_parse_options_from_toml_file(self):
+        with NamedTemporaryFile(mode="w", suffix=".toml") as file:
+            config_data = """\
+            [tool.locust]
+            locustfile = "./test_locustfile.py"
+            web-host = "127.0.0.1"
+            web-port = 45787
+            headless = true
+            tags = ["Critical", "Normal"]
+            """
+
+            file.write(config_data)
+            file.flush()
+            parser = get_parser(default_config_files=[file.name])
+            options = parser.parse_args(["-H", "https://example.com"])
+
+        self.assertEqual("./test_locustfile.py", options.locustfile)
+        self.assertEqual("127.0.0.1", options.web_host)
+        self.assertEqual(45787, options.web_port)
+        self.assertTrue(options.headless)
+        self.assertEqual(["Critical", "Normal"], options.tags)
+        self.assertEqual("https://example.com", options.host)
 
 
 class TestArgumentParser(LocustTestCase):
@@ -88,11 +117,39 @@ class TestArgumentParser(LocustTestCase):
         self.assertEqual(10, options.spawn_rate)
         self.assertEqual("5m", options.run_time)
         self.assertTrue(options.reset_stats)
-        self.assertEqual(5, options.stop_timeout)
+        self.assertEqual("5", options.stop_timeout)
         self.assertEqual(["MyUserClass"], options.user_classes)
         # check default arg
         self.assertEqual(8089, options.web_port)
 
+    def test_parse_options_from_env(self):
+        os.environ["LOCUST_LOCUSTFILE"] = "locustfile.py"
+        os.environ["LOCUST_USERS"] = "100"
+        os.environ["LOCUST_SPAWN_RATE"] = "10"
+        os.environ["LOCUST_RUN_TIME"] = "5m"
+        os.environ["LOCUST_RESET_STATS"] = "true"
+        os.environ["LOCUST_STOP_TIMEOUT"] = "5"
+        os.environ["LOCUST_USER_CLASSES"] = "MyUserClass"
+        options = parse_options(args=[])
+
+        self.assertEqual("locustfile.py", options.locustfile)
+        self.assertEqual(100, options.num_users)
+        self.assertEqual(10, options.spawn_rate)
+        self.assertEqual("5m", options.run_time)
+        self.assertTrue(options.reset_stats)
+        self.assertEqual("5", options.stop_timeout)
+        self.assertEqual(["MyUserClass"], options.user_classes)
+        # check default arg
+        self.assertEqual(8089, options.web_port)
+
+        del os.environ["LOCUST_LOCUSTFILE"]
+        del os.environ["LOCUST_USERS"]
+        del os.environ["LOCUST_SPAWN_RATE"]
+        del os.environ["LOCUST_RUN_TIME"]
+        del os.environ["LOCUST_RESET_STATS"]
+        del os.environ["LOCUST_STOP_TIMEOUT"]
+        del os.environ["LOCUST_USER_CLASSES"]
+
     def test_parse_locustfile(self):
         with mock_locustfile() as mocked:
             locustfiles = parse_locustfile_option(
@@ -286,15 +343,9 @@ class TestArgumentParser(LocustTestCase):
         def _(parser, **kw):
             parser.add_argument("--a1", help="a1 help")
             parser.add_argument("--a2", help="a2 help", include_in_web_ui=False)
+            parser.add_argument("--a3", help="a3 help", is_secret=True)
 
-        args = [
-            "-u",
-            "666",
-            "--a1",
-            "v1",
-            "--a2",
-            "v2",
-        ]
+        args = ["-u", "666", "--a1", "v1", "--a2", "v2", "--a3", "v3"]
         options = parse_options(args=args)
         self.assertEqual(666, options.num_users)
         self.assertEqual("v1", options.a1)
@@ -303,7 +354,8 @@ class TestArgumentParser(LocustTestCase):
         extra_args = ui_extra_args_dict(args)
         self.assertIn("a1", extra_args)
         self.assertNotIn("a2", extra_args)
-        self.assertEqual("v1", extra_args["a1"])
+        self.assertIn("a3", extra_args)
+        self.assertEqual("v1", extra_args["a1"]["default_value"])
 
 
 class TestFindLocustfiles(LocustTestCase):
