@@ -1,7 +1,21 @@
+from __future__ import annotations
+
+import locust
+from locust import LoadTestShape, constant, stats
+from locust.argument_parser import get_parser, parse_options
+from locust.env import Environment
+from locust.log import LogReader
+from locust.runners import Runner
+from locust.stats import StatsCSVFileWriter
+from locust.user import User, task
+from locust.web import WebUI
+
 import copy
 import csv
 import json
+import logging
 import os
+import re
 import textwrap
 import traceback
 from io import StringIO
@@ -9,22 +23,13 @@ from tempfile import NamedTemporaryFile, TemporaryDirectory
 
 import gevent
 import requests
+from flask_login import UserMixin
 from pyquery import PyQuery as pq
 
-import locust
-from locust import constant, LoadTestShape
-from locust.argument_parser import get_parser, parse_options
-from locust.user import User, task
-from locust.env import Environment
-from locust.runners import Runner
-from locust import stats
-from locust.stats import StatsCSVFileWriter
-from locust.web import WebUI
-
+from ..util.load_locustfile import load_locustfile
 from .mock_locustfile import mock_locustfile
 from .testcases import LocustTestCase
 from .util import create_tls_cert
-from ..util.load_locustfile import load_locustfile
 
 
 class _HeaderCheckMixin:
@@ -562,7 +567,15 @@ class TestWebUI(LocustTestCase, _HeaderCheckMixin):
             def t(self):
                 pass
 
-        class TestShape(LoadTestShape):
+        class TestShape1(LoadTestShape):
+            def tick(self):
+                run_time = self.get_run_time()
+                if run_time < 10:
+                    return 4, 4
+                else:
+                    return None
+
+        class TestShape2(LoadTestShape):
             def tick(self):
                 run_time = self.get_run_time()
                 if run_time < 10:
@@ -572,8 +585,8 @@ class TestWebUI(LocustTestCase, _HeaderCheckMixin):
 
         self.environment.web_ui.userclass_picker_is_active = True
         self.environment.available_user_classes = {"User1": User1, "User2": User2}
-        self.environment.available_shape_classes = {"TestShape": TestShape()}
-        self.environment.shape_class = None
+        self.environment.available_shape_classes = {"TestShape1": TestShape1(), "TestShape2": TestShape2()}
+        self.environment.shape_class = TestShape1()
 
         response = requests.post(
             "http://127.0.0.1:%i/swarm" % self.web_port,
@@ -582,14 +595,14 @@ class TestWebUI(LocustTestCase, _HeaderCheckMixin):
                 "spawn_rate": 5,
                 "host": "https://localhost",
                 "user_classes": "User1",
-                "shape_class": "TestShape",
+                "shape_class": "TestShape2",
             },
         )
 
         self.assertEqual(200, response.status_code)
         self.assertEqual("https://localhost", response.json()["host"])
         self.assertEqual(self.environment.host, "https://localhost")
-        assert isinstance(self.environment.shape_class, TestShape)
+        assert isinstance(self.environment.shape_class, TestShape2)
 
         # stop
         gevent.sleep(1)
@@ -646,6 +659,41 @@ class TestWebUI(LocustTestCase, _HeaderCheckMixin):
         response = requests.get("http://127.0.0.1:%i/stop" % self.web_port)
         self.assertEqual(response.json()["message"], "Test stopped")
 
+    def test_swarm_shape_class_is_updated_when_userclass_picker_is_active(self):
+        class User1(User):
+            pass
+
+        class TestShape(LoadTestShape):
+            def tick(self):
+                pass
+
+        test_shape_instance = TestShape()
+
+        self.environment.web_ui.userclass_picker_is_active = True
+        self.environment.available_user_classes = {"User1": User1}
+        self.environment.available_shape_classes = {"TestShape": test_shape_instance}
+        self.environment.shape_class = None
+
+        response = requests.post(
+            "http://127.0.0.1:%i/swarm" % self.web_port,
+            data={
+                "user_count": 5,
+                "spawn_rate": 5,
+                "host": "https://localhost",
+                "user_classes": "User1",
+                "shape_class": "TestShape",
+            },
+        )
+
+        self.assertEqual(200, response.status_code)
+        self.assertEqual(test_shape_instance, self.environment.shape_class)
+        self.assertIsNotNone(test_shape_instance.runner)
+
+        # stop
+        gevent.sleep(1)
+        response = requests.get("http://127.0.0.1:%i/stop" % self.web_port)
+        self.assertEqual(response.json()["message"], "Test stopped")
+
     def test_swarm_userclass_shapeclass_ignored_when_userclass_picker_is_inactive(self):
         class User1(User):
             wait_time = constant(1)
@@ -723,6 +771,30 @@ class TestWebUI(LocustTestCase, _HeaderCheckMixin):
         self.assertEqual(200, response.status_code)
         self.assertEqual(my_dict["val"], 42)
 
+    def test_custom_argument_dropdown(self):
+        class MyUser(User):
+            host = "http://example.com"
+
+        @locust.events.init_command_line_parser.add_listener
+        def _(parser, **kw):
+            parser.add_argument("--my-argument", default="a", choices=["a", "b"], help="Pick one")
+
+        parsed_options = parse_options(args=["--my-argument", "b"])
+        self.environment.user_classes = [MyUser]
+        self.environment.parsed_options = parsed_options
+        self.environment.web_ui.parsed_options = parsed_options
+
+        response = requests.get("http://127.0.0.1:%i/" % self.web_port)
+        self.assertEqual(200, response.status_code)
+
+        # regex to match the intended select tag with id from the custom argument
+        dropdown_pattern = re.compile(r"<select[^>]*id=\"my_argument\"[^>]*>", flags=re.I)
+        self.assertRegex(response.text, dropdown_pattern)
+
+        # regex to match the input that generates if it fails to find the choices
+        textfield_pattern = re.compile(r"<input[^>]*id=\"my_argument\"[^>]*>", flags=re.I)
+        self.assertNotRegex(response.text, textfield_pattern)
+
     def test_swarm_host_value_not_specified(self):
         class MyUser(User):
             wait_time = constant(1)
@@ -920,17 +992,149 @@ class TestWebUI(LocustTestCase, _HeaderCheckMixin):
             isinstance(next(iter(self.runner.exceptions.values()))["nodes"], set), "exception object has been mutated"
         )
 
+    def test_custom_shape_deactivate_num_users_and_spawn_rate(self):
+        class TestShape(LoadTestShape):
+            def tick(self):
+                return None
+
+        self.environment.shape_class = TestShape
+
+        response = requests.get("http://127.0.0.1:%i/" % self.web_port)
+        self.assertEqual(200, response.status_code)
+
+        # regex to match the intended select tag with id from the custom argument
+        re_disabled_user_count = re.compile(
+            r"<input[^>]*id=\"(new_)?user_count\"[^>]*disabled=\"disabled\"[^>]*>", flags=re.I
+        )
+        self.assertRegex(response.text, re_disabled_user_count)
+
+        re_disabled_spawn_rate = re.compile(
+            r"<input[^>]*id=\"(new_)?spawn_rate\"[^>]*disabled=\"disabled\"[^>]*>", flags=re.I
+        )
+        self.assertRegex(response.text, re_disabled_spawn_rate)
+
+    def test_custom_shape_with_use_common_options_keep_num_users_and_spawn_rate(self):
+        class TestShape(LoadTestShape):
+            use_common_options = True
+
+            def tick(self):
+                return None
+
+        self.environment.shape_class = TestShape
+
+        response = requests.get("http://127.0.0.1:%i/" % self.web_port)
+        self.assertEqual(200, response.status_code)
+
+        # regex to match the intended select tag with id from the custom argument
+        re_user_count = re.compile(r"<input[^>]*id=\"(new_)?user_count\"[^>]*>", flags=re.I)
+        re_disabled_user_count = re.compile(
+            r"<input[^>]*id=\"(new_)?user_count\"[^>]*disabled=\"disabled\"[^>]*>", flags=re.I
+        )
+        self.assertRegex(response.text, re_user_count)
+        self.assertNotRegex(response.text, re_disabled_user_count)
+
+        re_spawn_rate = re.compile(r"<input[^>]*id=\"(new_)?spawn_rate\"[^>]*>", flags=re.I)
+        re_disabled_spawn_rate = re.compile(
+            r"<input[^>]*id=\"(new_)?spawn_rate\"[^>]*disabled=\"disabled\"[^>]*>", flags=re.I
+        )
+        self.assertRegex(response.text, re_spawn_rate)
+        self.assertNotRegex(response.text, re_disabled_spawn_rate)
+
+    def test_html_stats_report(self):
+        self.environment.locustfile = "locust.py"
+        self.environment.host = "http://localhost"
+
+        response = requests.get("http://127.0.0.1:%i/stats/report" % self.web_port)
+        self.assertEqual(200, response.status_code)
+
+        d = pq(response.content.decode("utf-8"))
+
+        self.assertIn("Script: <span>locust.py</span>", str(d))
+        self.assertIn("Target Host: <span>http://localhost</span>", str(d))
+
+    def test_logs(self):
+        log_handler = LogReader()
+        log_handler.name = "log_reader"
+        log_handler.setLevel(logging.INFO)
+        logger = logging.getLogger("root")
+        logger.addHandler(log_handler)
+        log_line = "some log info"
+        logger.info(log_line)
+
+        response = requests.get("http://127.0.0.1:%i/logs" % self.web_port)
+
+        self.assertIn(log_line, response.json().get("logs"))
+
+    def test_template_args(self):
+        class MyUser(User):
+            @task
+            def do_something(self):
+                self.client.get("/")
+
+            host = "http://example.com"
+
+        class MyUser2(User):
+            host = "http://example.com"
+
+        self.environment.user_classes = [MyUser, MyUser2]
+        self.environment.available_user_classes = {"User1": MyUser, "User2": MyUser2}
+        self.environment.available_user_tasks = {"User1": MyUser.tasks, "User2": MyUser2.tasks}
+
+        users = {"User1": MyUser.json(), "User2": MyUser2.json()}
+        available_user_tasks = {"User1": ["do_something"], "User2": []}
+
+        self.web_ui.update_template_args()
+
+        self.assertEqual(self.web_ui.template_args.get("users"), users)
+        self.assertEqual(
+            self.web_ui.template_args.get("available_user_classes"), sorted(self.environment.available_user_classes)
+        )
+        self.assertEqual(self.web_ui.template_args.get("available_user_tasks"), available_user_tasks)
+
+    def test_update_user_endpoint(self):
+        class MyUser(User):
+            @task
+            def my_task(self):
+                pass
+
+            @task
+            def my_task_2(self):
+                pass
+
+            host = "http://example.com"
+
+        class MyUser2(User):
+            host = "http://example.com"
+
+        self.environment.user_classes = [MyUser, MyUser2]
+        self.environment.available_user_classes = {"User1": MyUser, "User2": MyUser2}
+        self.environment.available_user_tasks = {"User1": MyUser.tasks, "User2": MyUser2.tasks}
+
+        users = {"User1": MyUser.json(), "User2": MyUser2.json()}
+        available_user_tasks = {"User1": ["my_task", "my_task_2"], "User2": []}
+
+        # environment.update_user_class({"user_class_name": "User1", "host": "http://localhost", "tasks": ["my_task_2"]})
+        response = requests.post(
+            "http://127.0.0.1:%i/user" % self.web_port,
+            json={"user_class_name": "User1", "host": "http://localhost", "tasks": ["my_task_2"]},
+        )
+
+        self.assertEqual(
+            self.environment.available_user_classes["User1"].json(),
+            {"host": "http://localhost", "tasks": ["my_task_2"], "fixed_count": 0, "weight": 1},
+        )
+
 
 class TestWebUIAuth(LocustTestCase):
     def setUp(self):
         super().setUp()
 
         parser = get_parser(default_config_files=[])
-        options = parser.parse_args(["--web-auth", "john:doe"])
-        self.runner = Runner(self.environment)
-        self.stats = self.runner.stats
-        self.web_ui = self.environment.create_web_ui("127.0.0.1", 0, auth_credentials=options.web_auth)
-        self.web_ui.app.view_functions["request_stats"].clear_cache()
+        self.environment.parsed_options = parser.parse_args(["--web-login"])
+
+        self.web_ui = self.environment.create_web_ui("127.0.0.1", 0, modern_ui=True, web_login=True)
+
+        self.web_ui.app.secret_key = "secret!"
         gevent.sleep(0.01)
         self.web_port = self.web_ui.server.server_port
 
@@ -939,18 +1143,36 @@ class TestWebUIAuth(LocustTestCase):
         self.web_ui.stop()
         self.runner.quit()
 
-    def test_index_with_basic_auth_enabled_correct_credentials(self):
-        self.assertEqual(
-            200, requests.get("http://127.0.0.1:%i/?ele=phino" % self.web_port, auth=("john", "doe")).status_code
-        )
+    def test_index_with_web_login_enabled_valid_user(self):
+        class User(UserMixin):
+            def __init__(self):
+                self.username = "test_user"
 
-    def test_index_with_basic_auth_enabled_incorrect_credentials(self):
-        self.assertEqual(
-            401, requests.get("http://127.0.0.1:%i/?ele=phino" % self.web_port, auth=("john", "invalid")).status_code
-        )
+            def get_id(self):
+                return self.username
 
-    def test_index_with_basic_auth_enabled_blank_credentials(self):
-        self.assertEqual(401, requests.get("http://127.0.0.1:%i/?ele=phino" % self.web_port).status_code)
+        def load_user(id):
+            return User()
+
+        self.web_ui.login_manager.request_loader(load_user)
+
+        response = requests.get("http://127.0.0.1:%i" % self.web_port)
+        d = pq(response.content.decode("utf-8"))
+
+        self.assertNotIn("authArgs", str(d))
+        self.assertIn("templateArgs", str(d))
+
+    def test_index_with_web_login_enabled_no_user(self):
+        def load_user():
+            return None
+
+        self.web_ui.login_manager.user_loader(load_user)
+
+        response = requests.get("http://127.0.0.1:%i" % self.web_port)
+        d = pq(response.content.decode("utf-8"))
+
+        # asserts auth page is returned
+        self.assertIn("authArgs", str(d))
 
 
 class TestWebUIWithTLS(LocustTestCase):
@@ -995,6 +1217,7 @@ class TestWebUIWithTLS(LocustTestCase):
 
 
 class TestWebUIFullHistory(LocustTestCase, _HeaderCheckMixin):
+    STATS_BASE_DIR = "csv_output"
     STATS_BASE_NAME = "web_test"
     STATS_FILENAME = f"{STATS_BASE_NAME}_stats.csv"
     STATS_HISTORY_FILENAME = f"{STATS_BASE_NAME}_stats_history.csv"
@@ -1005,7 +1228,9 @@ class TestWebUIFullHistory(LocustTestCase, _HeaderCheckMixin):
         self.remove_files_if_exists()
 
         parser = get_parser(default_config_files=[])
-        self.environment.parsed_options = parser.parse_args(["--csv", self.STATS_BASE_NAME, "--csv-full-history"])
+        self.environment.parsed_options = parser.parse_args(
+            ["--csv", os.path.join(self.STATS_BASE_DIR, self.STATS_BASE_NAME), "--csv-full-history"]
+        )
         self.stats = self.environment.stats
         self.stats.CSV_STATS_INTERVAL_SEC = 0.02
 
@@ -1032,6 +1257,7 @@ class TestWebUIFullHistory(LocustTestCase, _HeaderCheckMixin):
         self.remove_file_if_exists(self.STATS_FILENAME)
         self.remove_file_if_exists(self.STATS_HISTORY_FILENAME)
         self.remove_file_if_exists(self.STATS_FAILURES_FILENAME)
+        self.remove_file_if_exists(self.STATS_BASE_DIR)
 
     def test_request_stats_full_history_csv(self):
         self.stats.log_request("GET", "/test", 1.39764125, 2)
@@ -1058,3 +1284,70 @@ class TestWebUIFullHistory(LocustTestCase, _HeaderCheckMixin):
         self.assertEqual("/test2", rows[2][3])
         self.assertEqual("", rows[3][2])
         self.assertEqual("Aggregated", rows[3][3])
+
+
+class TestModernWebUI(LocustTestCase, _HeaderCheckMixin):
+    def setUp(self):
+        super().setUp()
+
+        parser = get_parser(default_config_files=[])
+        self.stats = self.environment.stats
+
+        self.web_ui = self.environment.create_web_ui("127.0.0.1", 0, modern_ui=True)
+        self.web_ui.app.view_functions["request_stats"].clear_cache()
+        gevent.sleep(0.01)
+        self.web_port = self.web_ui.server.server_port
+
+    def tearDown(self):
+        super().tearDown()
+        self.web_ui.stop()
+        self.runner.quit()
+
+    def test_index_with_modern_ui(self):
+        self.assertEqual(self.web_ui, self.environment.web_ui)
+
+        html_to_option = {
+            "num_users": ["-u", "100"],
+            "spawn_rate": ["-r", "10.0"],
+        }
+
+        response = requests.get("http://127.0.0.1:%i/" % self.web_port)
+        d = pq(response.content.decode("utf-8"))
+
+        self.assertEqual(200, response.status_code)
+        self.assertTrue(d("#root"))
+
+        for html_name_to_test in html_to_option.keys():
+            # Test that setting each spawn option individually populates the corresponding field in the html, and none of the others
+            self.environment.parsed_options = parse_options(html_to_option[html_name_to_test])
+
+            response = requests.get("http://127.0.0.1:%i/" % self.web_port)
+            self.assertEqual(200, response.status_code)
+
+            d = pq(response.content.decode("utf-8"))
+
+            self.assertIn(f'"{html_name_to_test}": {html_to_option[html_name_to_test][1]}', str(d("script")))
+
+    def test_web_ui_no_runner(self):
+        env = Environment()
+        web_ui = WebUI(env, "127.0.0.1", 0)
+        gevent.sleep(0.01)
+        try:
+            response = requests.get("http://127.0.0.1:%i/" % web_ui.server.server_port)
+            self.assertEqual(500, response.status_code)
+            self.assertEqual("Error: Locust Environment does not have any runner", response.text)
+        finally:
+            web_ui.stop()
+
+    def test_html_stats_report(self):
+        self.environment.locustfile = "locust.py"
+        self.environment.host = "http://localhost"
+
+        response = requests.get("http://127.0.0.1:%i/stats/report" % self.web_port)
+        self.assertEqual(200, response.status_code)
+
+        d = pq(response.content.decode("utf-8"))
+
+        self.assertTrue(d("#root"))
+        self.assertIn('"locustfile": "locust.py"', str(d))
+        self.assertIn('"host": "http://localhost"', str(d))
